"""Classes/methods for loading YAML intent files."""

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import IO, Any, Dict, Iterable, List, Tuple

from dataclasses_json import dataclass_json
from yaml import safe_load

from .expression import Sentence
from .parse import parse_sentence, parse_sentences


class ResponseType(str, Enum):
    SUCCESS = "success"
    NO_INTENT = "no_intent"
    NO_AREA = "no_area"
    NO_DOMAIN = "no_domain"
    NO_DEVICE_CLASS = "no_device_class"
    NO_ENTITY = "no_entity"
    HANDLE_ERROR = "handle_error"


@dataclass
class IntentData:
    """Block of sentences and known slots for an intent."""

    sentences: List[Sentence]
    slots: Dict[str, Any] = field(default_factory=dict)


@dataclass_json
@dataclass
class Intent:
    """A named intent with sentences + slots."""

    name: str
    data: List[IntentData] = field(default_factory=list)


class SlotList(ABC):
    """Base class for slot lists."""


class RangeType(str, Enum):
    """Number range type."""

    NUMBER = "number"
    PERCENTAGE = "percentage"
    TEMPERATURE = "temperature"


@dataclass
class RangeSlotList(SlotList):
    """Slot list for a range of numbers."""

    start: int
    stop: int
    step: int = 1
    type: RangeType = RangeType.NUMBER

    def __post_init__(self):
        """Validate number range"""
        assert self.start < self.stop, "start must be less than stop"
        assert self.step > 0, "step must be positive"


@dataclass_json
@dataclass
class TextSlotValue:
    """Single value in a text slot list."""

    text_in: Sentence
    """Input text for this value"""

    value_out: Any
    """Output value put into slot"""


@dataclass
class TextSlotList(SlotList):
    """Slot list with pre-defined text values."""

    values: List[TextSlotValue]

    @staticmethod
    def from_strings(strings: Iterable[str]) -> "TextSlotList":
        """
        Construct a text slot list from strings.

        Input and output values are the same text.
        """
        return TextSlotList(
            values=[
                TextSlotValue(
                    text_in=parse_sentence(text, keep_text=True), value_out=text
                )
                for text in strings
            ],
        )

    @staticmethod
    def from_tuples(tuples: Iterable[Tuple[str, Any]]) -> "TextSlotList":
        """
        Construct a text slot list from text/value pairs.

        Input values are the left (text), output values are the right (any).
        """
        return TextSlotList(
            values=[
                TextSlotValue(
                    text_in=parse_sentence(text, keep_text=True), value_out=value
                )
                for text, value in tuples
            ],
        )


@dataclass
class Intents:
    """Collection of intents, rules, and lists for a language."""

    language: str
    """Language code (e.g., en)."""

    intents: Dict[str, Intent]
    """Intents mapped by name."""

    slot_lists: Dict[str, SlotList] = field(default_factory=dict)
    """Slot lists mapped by name."""

    expansion_rules: Dict[str, Sentence] = field(default_factory=dict)
    """Expansion rules mapped by name."""

    skip_words: List[str] = field(default_factory=list)
    """Words that can be skipped during recognition."""

    @staticmethod
    def from_yaml(yaml_file: IO[str]) -> "Intents":
        """Load intents from a YAML file."""
        return Intents.from_dict(safe_load(yaml_file))

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> "Intents":
        """Parse intents from a dict."""
        # language: "<code>"
        # intents:
        #   IntentName:
        #     data:
        #       - sentences:
        #           - "<sentence>"
        #         slots:
        #           <slot_name>: <slot value>
        #           <slot_name>:
        #             - <slot value>
        # expansion_rules:
        #   <rule_name>: "<rule body>"
        # lists:
        #   <list_name>:
        #     values:
        #       - "<value>"
        #
        return Intents(
            language=input_dict["language"],
            intents={
                intent_name: Intent(
                    name=intent_name,
                    data=[
                        IntentData(
                            sentences=parse_sentences(
                                data_dict["sentences"], keep_text=True
                            ),
                            slots=data_dict.get("slots", {}),
                        )
                        for data_dict in intent_dict["data"]
                    ],
                )
                for intent_name, intent_dict in input_dict["intents"].items()
            },
            slot_lists={
                list_name: _parse_list(list_dict)
                for list_name, list_dict in input_dict.get("lists", {}).items()
            },
            expansion_rules={
                rule_name: parse_sentences([rule_body], keep_text=True)[0]
                for rule_name, rule_body in input_dict.get(
                    "expansion_rules", {}
                ).items()
            },
            skip_words=input_dict.get("skip_words", []),
        )


def _parse_list(list_dict: Dict[str, Any]) -> SlotList:
    """Parses a slot list from a dict."""
    if "values" in list_dict:
        # Text values
        text_values: List[TextSlotValue] = []
        for value in list_dict["values"]:
            if isinstance(value, str):
                # String value
                text_values.append(
                    TextSlotValue(
                        text_in=parse_sentence(value, keep_text=True), value_out=value
                    )
                )
            else:
                # Object with "in" and "out"
                text_values.append(
                    TextSlotValue(
                        text_in=parse_sentence(value["in"], keep_text=True),
                        value_out=value["out"],
                    )
                )

        return TextSlotList(text_values)

    if "range" in list_dict:
        # Number range
        range_dict = list_dict["range"]
        return RangeSlotList(
            type=RangeType(range_dict.get("type", "number")),
            start=int(range_dict["from"]),
            stop=int(range_dict["to"]),
            step=int(range_dict.get("step", 1)),
        )

    raise ValueError(f"Unknown slot list type: {list_dict}")



import io

import pytest

from hassil import Intents, recognize
from hassil.intents import TextSlotList

TEST_YAML = """
language: "en"
intents:
  TurnOnTV:
    data:
      - sentences:
        - "turn on [the] TV in <area>"
        - "turn on <area> TV"
        slots:
          domain: "media_player"
          name: "roku"
  SetBrightness:
    data:
      - sentences:
        - "set [the] brightness in <area> to <brightness>"
        slots:
          domain: "light"
          name: "all"
  GetTemperature:
    data:
      - sentences:
        - "<what_is> [the] temperature in <area>"
        slots:
          domain: "climate"
expansion_rules:
  area: "[the] {area}"
  brightness: "{brightness_pct} [percent]"
  what_is: "(what's | whats | what is)"
lists:
  brightness_pct:
    range:
      type: percentage
      from: 0
      to: 100
skip_words:
  - "please"
"""


@pytest.fixture
def intents():
    with io.StringIO(TEST_YAML) as test_file:
        return Intents.from_yaml(test_file)


@pytest.fixture
def slot_lists():
    return {
        "area": TextSlotList.from_tuples(
            [("kitchen", "area.kitchen"), ("living room", "area.living_room")]
        )
    }


# pylint: disable=redefined-outer-name
def test_turn_on(intents, slot_lists):
    result = recognize("turn on kitchen TV, please", intents, slot_lists=slot_lists)
    assert result is not None
    assert result.intent.name == "TurnOnTV"

    area = result.entities["area"]
    assert area.name == "area"
    assert area.value == "area.kitchen"

    # From YAML
    assert result.entities["domain"].value == "media_player"
    assert result.entities["name"].value == "roku"


# pylint: disable=redefined-outer-name
def test_brightness(intents, slot_lists):
    result = recognize(
        "set the brightness in the living room to 75%", intents, slot_lists=slot_lists
    )
    assert result is not None
    assert result.intent.name == "SetBrightness"

    assert result.entities["area"].value == "area.living_room"
    assert result.entities["brightness_pct"].value == 75

    # From YAML
    assert result.entities["domain"].value == "light"
    assert result.entities["name"].value == "all"


# pylint: disable=redefined-outer-name
def test_temperature(intents, slot_lists):
    result = recognize(
        "what is the temperature in the living room?", intents, slot_lists=slot_lists
    )
    assert result is not None
    assert result.intent.name == "GetTemperature"

    assert result.entities["area"].value == "area.living_room"

    # From YAML
    assert result.entities["domain"].value == "climate"



"""Methods for recognizing intents from text."""

import dataclasses
import itertools
import re
from dataclasses import dataclass, field
from os.path import commonprefix
from typing import Any, Dict, Iterable, List, Optional, Tuple

from .expression import (
    Expression,
    ListReference,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    TextChunk,
)
from .intents import Intent, Intents, RangeSlotList, SlotList, TextSlotList
from .util import normalize_text

NUMBER_START = re.compile(r"^(\s*-?[0-9]+)")

NON_WORD = re.compile(r"^(\W+\s*)")


class MissingListError(Exception):
    """Error when a {slot_list} is missing."""


class MissingRuleError(Exception):
    """Error when an <expansion_rule> is missing."""


@dataclass
class MatchEntity:
    """Named entity that has been matched from a {slot_list}"""

    name: str
    """Name of the entity."""

    value: Any
    """Value of the entity."""


@dataclass
class MatchContext:
    """Context passed to match_expression."""

    text: str
    """Input text remaining to be processed."""

    text_index: int = 0
    """Index of text currently being processed."""

    slot_lists: Dict[str, SlotList] = field(default_factory=dict)
    """Available slot lists mapped by name."""

    expansion_rules: Dict[str, Sentence] = field(default_factory=dict)
    """Available expansion rules mapped by name."""

    skip_words: List[re.Pattern] = field(default_factory=list)
    """Words that can be skipped during recognition."""

    entities: List[MatchEntity] = field(default_factory=list)
    """Entities that have been found in input text."""

    @property
    def is_match(self) -> bool:
        """True if no words are left"""
        return not self.text


@dataclass
class RecognizeResult:
    """Result of recognition."""

    intent: Intent
    """Matched intent"""

    entities: Dict[str, MatchEntity] = field(default_factory=dict)
    """Matched entities mapped by name."""

    entities_list: List[MatchEntity] = field(default_factory=list)
    """Matched entities as a list (duplicates allowed)."""


def recognize(
    text: str,
    intents: Intents,
    slot_lists: Optional[Dict[str, SlotList]] = None,
    expansion_rules: Optional[Dict[str, Sentence]] = None,
    skip_words: Optional[Iterable[str]] = None,
) -> Optional[RecognizeResult]:
    """Return the first match of input text/words against a collection of intents."""
    text = normalize_text(text)

    if slot_lists is None:
        slot_lists = intents.slot_lists
    else:
        # Combine with intents
        slot_lists = {**intents.slot_lists, **slot_lists}

    if slot_lists is None:
        slot_lists = {}

    if expansion_rules is None:
        expansion_rules = intents.expansion_rules
    else:
        # Combine rules
        expansion_rules = {**intents.expansion_rules, **expansion_rules}

    if skip_words is None:
        skip_words = intents.skip_words
    else:
        # Combine skip words
        skip_words = itertools.chain(skip_words, intents.skip_words)

    # Check sentence against each intent.
    # This should eventually be done in parallel.
    for intent in intents.intents.values():
        for intent_data in intent.data:
            for intent_sentence in intent_data.sentences:
                # Create initial context
                context = MatchContext(
                    text=text,
                    slot_lists=slot_lists,
                    expansion_rules=expansion_rules,
                    skip_words=_prepare_skip_words(skip_words),
                )
                sentence_contexts = _match_and_skip(context, intent_sentence)
                for sentence_context in sentence_contexts:
                    if sentence_context.is_match:
                        # Add fixed entities
                        for slot_name, slot_value in intent_data.slots.items():
                            sentence_context.entities.append(
                                MatchEntity(name=slot_name, value=slot_value)
                            )

                        # Return the first match
                        return RecognizeResult(
                            intent,
                            {
                                entity.name: entity
                                for entity in sentence_context.entities
                            },
                            sentence_context.entities,
                        )

    return None


def is_match(
    text: str,
    sentence: Sentence,
    slot_lists: Optional[Dict[str, SlotList]] = None,
    expansion_rules: Optional[Dict[str, Sentence]] = None,
    skip_words: Optional[Iterable[str]] = None,
    entities: Optional[Dict[str, Any]] = None,
) -> Optional[MatchContext]:
    """Return the first match of input text/words against a sentence expression."""
    text = normalize_text(text)

    if slot_lists is None:
        slot_lists = {}

    if expansion_rules is None:
        expansion_rules = {}

    if skip_words is None:
        skip_words = []
    else:
        skip_words = sorted(skip_words, key=len, reverse=True)

    context = MatchContext(
        text=text,
        slot_lists=slot_lists,
        expansion_rules=expansion_rules,
        skip_words=_prepare_skip_words(skip_words),
    )

    for match_context in _match_and_skip(context, sentence):
        if match_context.is_match:
            return match_context

    return None


def _prepare_skip_words(skip_words: Iterable[str]) -> List[re.Pattern]:
    """Convert skip word strings to regex patterns."""
    patterns: List[re.Pattern] = []
    for skip_word in skip_words:
        skip_word = normalize_text(skip_word)
        patterns.append(re.compile(rf"^{re.escape(skip_word)}\b"))

    return patterns


def _match_and_skip(context: MatchContext, *args, **kwargs) -> Iterable[MatchContext]:
    """Match a sentence, then skip any skippable words at the end of input"""
    for match_context in match_expression(context, *args, *kwargs):
        while match_context.text:
            can_skip, match_context.text = _skip(match_context.text, context.skip_words)
            if not can_skip:
                break

        yield match_context


def match_expression(
    context: MatchContext, expression: Expression
) -> Iterable[MatchContext]:
    """Yield matching contexts for an expresion"""
    if isinstance(expression, TextChunk):
        chunk: TextChunk = expression
        if chunk.is_empty:
            # Skip template chunk
            yield context

        context_text_left = context.text
        chunk_text_left = chunk.text

        while chunk_text_left:
            prefix = commonprefix((context_text_left, chunk_text_left))

            if prefix:
                # Common prefix
                context_text_left = context_text_left[len(prefix) :]
                chunk_text_left = chunk_text_left[len(prefix) :]
            else:
                can_skip, context_text_left = _skip(
                    context_text_left, context.skip_words
                )

                if not can_skip:
                    break

        if not chunk_text_left:
            yield dataclasses.replace(context, text=context_text_left)

    elif isinstance(expression, Sequence):
        seq: Sequence = expression
        if seq.type == SequenceType.ALTERNATIVE:
            # Any may match (words | in | alternative)
            for item in seq.items:
                yield from match_expression(context, item)

        elif seq.type == SequenceType.GROUP:
            # All must match (words in group)
            # NOTE: [optional] = (optional | )
            if seq.items:
                group_contexts = [context]
                for item in seq.items:
                    # Next step
                    group_contexts = [
                        item_context
                        for group_context in group_contexts
                        for item_context in match_expression(group_context, item)
                    ]
                    if not group_contexts:
                        break

                for group_context in group_contexts:
                    yield group_context
        else:
            raise ValueError(f"Unexpected sequence type: {seq}")

    elif isinstance(expression, ListReference):
        # {list}
        list_ref: ListReference = expression
        if (not context.slot_lists) or (list_ref.list_name not in context.slot_lists):
            raise MissingListError(f"Missing slot list {{{list_ref.list_name}}}")

        slot_list = context.slot_lists[list_ref.list_name]
        if isinstance(slot_list, TextSlotList):
            text_list: TextSlotList = slot_list

            if context.text:
                # Any value may match
                for slot_value in text_list.values:
                    value_contexts = match_expression(
                        dataclasses.replace(context), slot_value.text_in
                    )

                    for value_context in value_contexts:
                        value_context.entities.append(
                            MatchEntity(
                                name=list_ref.slot_name, value=slot_value.value_out
                            )
                        )
                        yield value_context

        elif isinstance(slot_list, RangeSlotList):
            # List that represents a number range.
            # Numbers must currently be digits ("1" not "one").
            range_list: RangeSlotList = slot_list
            if context.text:
                number_match = NUMBER_START.match(context.text)
                if number_match is not None:
                    number_text = number_match[1]
                    word_number = int(number_text)
                    if range_list.step == 1:
                        # Unit step
                        in_range = range_list.start <= word_number <= range_list.stop
                    else:
                        # Non-unit step
                        in_range = word_number in range(
                            range_list.start, range_list.stop + 1, range_list.step
                        )

                    if in_range:
                        context.entities.append(
                            MatchEntity(name=list_ref.slot_name, value=word_number)
                        )

                        yield dataclasses.replace(
                            context, text=context.text[len(number_text) :].lstrip()
                        )

        else:
            raise ValueError(f"Unexpected slot list type: {slot_list}")

    elif isinstance(expression, RuleReference):
        # <rule>
        rule_ref: RuleReference = expression
        if (not context.expansion_rules) or (
            rule_ref.rule_name not in context.expansion_rules
        ):
            raise MissingRuleError(f"Missing expansion rule <{rule_ref.rule_name}>")

        yield from match_expression(
            context, context.expansion_rules[rule_ref.rule_name]
        )
    else:
        raise ValueError(f"Unexpected expression: {expression}")


def _skip(text: str, skip_words: Iterable[re.Pattern]) -> Tuple[bool, str]:
    """Skip over skip/non-words."""
    # Skip words
    for skip_word in skip_words:
        skip_word_match = skip_word.match(text)
        if skip_word_match is not None:
            text = text[len(skip_word_match[0]) :].lstrip()
            return True, text

    # Non-words
    nonword_match = NON_WORD.match(text)
    if nonword_match is not None:
        # Skip nonword parts
        nonword_text = nonword_match[1]
        text = text[len(nonword_text) :]
        return True, text

    return False, text



"""Utility methods"""
import collections
import re
import unicodedata


def merge_dict(base_dict, new_dict):
    """Merges new_dict into base_dict."""
    for key, value in new_dict.items():
        if key in base_dict:
            old_value = base_dict[key]
            if isinstance(old_value, collections.abc.MutableMapping):
                # Combine dictionary
                assert isinstance(
                    value, collections.abc.Mapping
                ), f"Not a dict: {value}"
                merge_dict(old_value, value)
            elif isinstance(old_value, collections.abc.MutableSequence):
                # Combine list
                assert isinstance(
                    value, collections.abc.Sequence
                ), f"Not a list: {value}"
                old_value.extend(value)
            else:
                # Overwrite
                base_dict[key] = value
        else:
            base_dict[key] = value


def remove_escapes(text: str) -> str:
    """Remove backslash escape sequences."""
    return re.sub(r"\\(.)", r"\1", text)


def remove_quotes(text: str) -> str:
    """Remove double quotes."""
    if (len(text) > 1) and (text[0] == '"') and (text[-1] == '"'):
        text = text[1:-1]

    return text


def normalize_whitespace(text: str) -> str:
    """Makes all whitespace inside a string single spaced."""
    return " ".join(text.split())


def normalize_text(text: str) -> str:
    """Normalize whitespace and unicode forms."""
    text = normalize_whitespace(text)
    text = text.casefold()
    text = unicodedata.normalize("NFC", text)

    return text



"""Tests for HassILExpressionListener"""
from hassil import (
    ListReference,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    TextChunk,
    parse_sentences,
)


def test_just_words():
    assert parse_sentences(["this is a test"]) == [s([tc("this is a test")])]


def test_group():
    assert parse_sentences(["(this is a test)"]) == [
        s(
            [
                grp(
                    [tc("this is a test")],
                )
            ]
        )
    ]


def test_optional():
    assert parse_sentences(["this is [a] test"]) == [
        s(
            [
                tc("this is"),
                alt(
                    [tc("a"), TextChunk.empty()],
                ),
                tc("test"),
            ]
        )
    ]


def test_optional_nested():
    assert parse_sentences(["this [is [a]] test"]) == [
        s(
            [
                tc("this"),
                alt(
                    [
                        grp([tc("is"), alt([tc("a"), TextChunk.empty()])]),
                        TextChunk.empty(),
                    ]
                ),
                tc("test"),
            ]
        )
    ]


def test_alternative():
    assert parse_sentences(["this is (a | the) test"]) == [
        s(
            [
                tc("this is"),
                alt(
                    [grp([tc("a")]), grp([tc("the")])],
                ),
                tc("test"),
            ]
        )
    ]


def test_alternative_multiple_words():
    assert parse_sentences(["this is (a bigger | the biggest) test"]) == [
        s(
            [
                tc("this is"),
                alt(
                    [
                        grp(
                            [tc("a bigger")],
                        ),
                        grp(
                            [tc("the biggest")],
                        ),
                    ],
                ),
                tc("test"),
            ]
        )
    ]


def test_alternative_what():
    assert parse_sentences(["(what | what's | whats | what is)"]) == [
        s(
            [
                alt(
                    [
                        grp(
                            [tc("what")],
                        ),
                        grp(
                            [tc("what's")],
                        ),
                        grp(
                            [tc("whats")],
                        ),
                        grp(
                            [tc("what is")],
                        ),
                    ],
                ),
            ]
        )
    ]


def test_list_reference():
    assert parse_sentences(["this is a {test}"]) == [
        s([tc("this is a"), ListReference("test")])
    ]


def test_list_reference_prefix_suffix():
    assert parse_sentences(["this is a pre'{test}-post"]) == [
        s([tc("this is a pre'"), ListReference("test"), tc("-post")])
    ]


def test_rule_reference():
    assert parse_sentences(["this is a <test>"]) == [
        s([tc("this is a"), RuleReference("test")])
    ]


def test_escape():
    assert parse_sentences(["this \\[is\\] a \\{test\\}"]) == [
        s([tc("this [is] a {test}")])
    ]


# -----------------------------------------------------------------------------


def s(items):
    return Sentence(items=items)


def tc(text):
    return TextChunk(text)


def grp(items):
    return Sequence(type=SequenceType.GROUP, items=items)


def alt(items):
    return Sequence(type=SequenceType.ALTERNATIVE, items=items)



"""Command-line interface to hassil."""
import argparse
import logging
import os
import sys
from pathlib import Path

import yaml

from .intents import Intents, TextSlotList
from .recognize import recognize
from .util import merge_dict

_LOGGER = logging.getLogger("hassil")


def main():
    """Main entry point"""
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml", nargs="+", help="YAML files or directories")
    parser.add_argument(
        "--areas",
        nargs="+",
        help="Area names",
        default=["kitchen", "bedroom", "living room"],
    )
    parser.add_argument("--names", nargs="+", default=[], help="Device/entity names")
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    args = parser.parse_args()

    level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(level=level)
    _LOGGER.debug(args)

    slot_lists = {
        "area": TextSlotList.from_strings(args.areas),
        "name": TextSlotList.from_strings(args.names),
    }

    input_dict = {"intents": {}}
    for yaml_path_str in args.yaml:
        yaml_path = Path(yaml_path_str)
        if yaml_path.is_dir():
            yaml_file_paths = yaml_path.glob("*.yaml")
        else:
            yaml_file_paths = [yaml_path]

        for yaml_file_path in yaml_file_paths:
            _LOGGER.debug("Loading file: %s", yaml_file_path)
            with open(yaml_file_path, "r", encoding="utf-8") as yaml_file:
                merge_dict(input_dict, yaml.safe_load(yaml_file))

    assert input_dict, "No intent YAML files loaded"
    intents = Intents.from_dict(input_dict)

    _LOGGER.info("Area names: %s", args.areas)
    _LOGGER.info("Device/Entity names: %s", args.names)

    if os.isatty(sys.stdout.fileno()):
        print("Reading sentences from stdin...", file=sys.stderr)

    try:
        for line in sys.stdin:
            line = line.strip()
            if not line:
                continue

            try:
                result = recognize(line, intents, slot_lists=slot_lists)
                if result is not None:
                    print(
                        {
                            "intent": result.intent.name,
                            **{e.name: e.value for e in result.entities_list},
                        }
                    )
                else:
                    print("<no match>")
            except Exception:
                _LOGGER.exception(line)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""Shared access to package resources"""
import os
import typing
from pathlib import Path

try:
    import importlib.resources

    files = importlib.resources.files
except (ImportError, AttributeError):
    # Backport for Python < 3.9
    import importlib_resources  # type: ignore

    files = importlib_resources.files

_PACKAGE = "hassil"
_DIR = Path(typing.cast(os.PathLike, files(_PACKAGE)))

__version__ = (_DIR / "VERSION").read_text(encoding="utf-8").strip()
